#pragma once

/** @file Demodulation.h
	@brief Contains implementations of down-conversion/devcimation algorithms
	using AVX2 and AVX512
 *****************************************************************************/

/** @defgroup demodavx Vectorized demodulation functions
	These functions implement down-conversion and decimation filters 
	using vector arithmetics based on AVX2 and AVX512.
 *****************************************************************************/
/** @{ */

#include "BiquadDef.h"

#ifndef __cplusplus
#include <stdalign.h>
#endif

/// Maximum stride length
#define MAX_STRIDE 32

#ifdef __cplusplus
extern "C" {
#endif

	/// Structure for demodulation decimation storing history and static data
	/// one rotation, mutiple inputs
	/// 
	/// Allocate one structure for each unique call (each data set) to 
	/// demodulation_decimation_stride8_section3. Call avx_stride_struct_init
	/// immediately afterwards.
	typedef struct {
		/// Filter history
		alignas(64) double hist[4 * MAX_STRIDE * 3];
		///  Old rotation stage [cos, sin]
		alignas(16) double rotation[2];
		/// Cordic angle [cos, sin]
		alignas(16) double anglexy[2];
		/// 2 * pi / (data rate); used to determine the cordic angel from the frequency
		double two_pi_over_rate;
		/// Old frequency used to compute the cordic angle
		double freqhist;
		/// counter to slow down old frequency comparison
		unsigned int pause;
	} avx_stride_struct;

	/// Initialization of the avx_stride_struct
	/// 
	/// @param init pointer to structure avx_stride_struct
	/// @param rate Sampling rate
	void avx_stride_struct_init(avx_stride_struct* init, double rate);

	/// Sets the frequency of the rotation avx_stride_struct
	/// 
	/// @param init pointer to structure avx_stride_struct
	/// @param freq Frequency of rotation
	void avx_stride_struct_set_frequency(avx_stride_struct* self, double freq);

	/// Applies one rotation step using the Cordic algorithm to the avx_stride_struct
	/// 
	/// @param init pointer to structure avx_stride_struct
	void avx_stride_struct_cordic(avx_stride_struct* self);

	/// Structure for demodulation decimation storing history and static data
	/// multiple rotation, one input
	typedef struct {
		/// Filter histroy
		alignas(64) double hist[4 * MAX_STRIDE * 3];
		///  Old rotation stage [cos, sin]
		alignas(64) double rotation[2 * MAX_STRIDE];
		/// Cordic angle [cos, sin]
		alignas(64) double anglexy[2 * MAX_STRIDE];
		/// Old frequency used to compute the cordic angle
		alignas(64) double freqhist[MAX_STRIDE];
		/// 2 * pi / (data rate); used to determine the cordic angel from the frequency
		double two_pi_over_rate;
		/// counter to slow down old frequency comparison
		unsigned int pause;
		/// index to slow down old frequency comparison
		unsigned int idx;
	} avx_rotation_struct;

	/// Initialization of avx_rotation_struct
	/// @param init pointer to structure
	/// @param rate Sampling rate
	void avx_rotation_struct_init(avx_rotation_struct* init, double rate);

	/// Sets the frequency of the rotation avx_rotation_struct
	/// 
	/// @param init pointer to structure avx_stride_struct
	/// @param freq Frequency of rotation
	/// @param idx Index into frequency array
	void avx_rotation_struct_set_frequency(avx_rotation_struct* self, double freq, unsigned int idx);


	/// Demodulation decimation on an array of input values using a single 
	/// down-conversion frequency.
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded second-order sections that form an IIR decimation filter. This 
	/// function doesn't have an intrinsic restriction on the number of filter 
	/// sections and input strides, but the corresponding AVX2 and AVX512 routines 
	/// only support strides of 8, 16, 24 or 32, and a filter of order 6th 
	/// (3 second order sections). So, this is enforced here too to be compatible.
	/// 
	/// @param inp Input values arranged as an array of length stride 
	/// @param freq Frequency of down-conversion
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_stride8_section3_std(const double* inp, double freq,
		double* out, const double* coeff, avx_stride_struct* hist, size_t stride);

	/// Demodulation decimation on an input value using multiple down-conversion
	/// frequencies.
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded second-order sections that form an IIR decimation filter. This 
	/// function doesn't have an intrinsic restriction on the number of filter 
	/// sections and input strides, but the corresponding AVX2 and AVX512 routines 
	/// only support strides of 8, 16, 24 or 32, and a filter of order 6th 
	/// (3 second order sections). So, this is enforced here too to be compatible.
	/// 
	/// @param inp Input value
	/// @param freq Frequencies for down-conversion arranged in an array of length stride
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_rotation8_section3_std(double inp, const double* freq,
		double* out, const double* coeff, avx_rotation_struct* hist, size_t stride);

	/// Demodulation decimation on an array of input values using a single 
	/// down-conversion frequency. 
	/// 
	/// SSE2/FMA version.
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded  second-order sections that form an IIR decimation filter. This 
	/// function has a restriction on the number of filter sections and input 
	/// strides, it only support strides of 8, 16, 24 or 32 and a filter of 
	/// order 6th (3 second order sections).
	/// 
	/// @param inp Input values arranged as an array of length stride 
	/// @param freq Frequency of down-conversion
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_stride8_section3_sse3(const double* inp, double freq,
		double* out, const double* coeff, avx_stride_struct* hist, size_t stride);

	/// Demodulation decimation on an input value using multiple down-conversion
	/// frequencies.
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded  second-order sections that form an IIR decimation filter. This 
	/// function has a restriction on the number of filter sections and input 
	/// strides, it only support strides of 8, 16, 24 or 32 and a filter of 
	/// order 6th (3 second order sections).
	/// 
	/// SSE2/FMA version.
	/// 
	/// @param inp Input value
	/// @param freq Frequencies for down-conversion arranged in an array of length stride
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_rotation8_section3_sse3(double inp, const double* freq,
		double* out, const double* coeff, avx_rotation_struct* hist, size_t stride);



	/// Demodulation decimation on an array of input values using a single 
	/// down-conversion frequency. 
	/// 
	/// AVX2 version.
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded  second-order sections that form an IIR decimation filter. This 
	/// function has a restriction on the number of filter sections and input 
	/// strides, it only support strides of 8, 16, 24 or 32 and a filter of 
	/// order 6th (3 second order sections).
	/// 
	/// @param inp Input values arranged as an array of length stride 
	/// @param freq Frequency of down-conversion
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_stride8_section3_avx2(const double* inp, double freq,
		double* out, const double* coeff, avx_stride_struct* hist, size_t stride);

	/// Demodulation decimation on an input value using multiple down-conversion
	/// frequencies.
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded  second-order sections that form an IIR decimation filter. This 
	/// function has a restriction on the number of filter sections and input 
	/// strides, it only support strides of 8, 16, 24 or 32 and a filter of 
	/// order 6th (3 second order sections).
	/// 
	/// AVX2 version.
	/// 
	/// @param inp Input value
	/// @param freq Frequencies for down-conversion arranged in an array of length stride
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_rotation8_section3_avx2(double inp, const double* freq,
		double* out, const double* coeff, avx_rotation_struct* hist, size_t stride);



	/// Demodulation decimation on an array of input values using a single down-conversion
	/// frequency. 
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded  second-order sections that form an IIR decimation filter. This 
	/// function has a restriction on the number of filter sections and input 
	/// strides, it only support strides of 8, 16, 24 or 32 and a filter of 
	/// order 6th (3 second order sections).
	/// 
	/// AVX512 version.
	/// 
	/// @param inp Input values arranged as an array of length stride 
	/// @param freq Frequency of down-conversion
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_stride8_section3_avx512(const double* inp, double freq,
		double* out, const double* coeff, avx_stride_struct* hist, size_t stride);

	/// Demodulation decimation on an input value using multiple down-conversion
	/// frequencies.
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded  second-order sections that form an IIR decimation filter. This 
	/// function has a restriction on the number of filter sections and input 
	/// strides, it only support strides of 8, 16, 24 or 32 and a filter of 
	/// order 6th (3 second order sections).
	/// 
	/// AVX512 version.
	/// 
	/// @param inp Input value
	/// @param freq Frequencies for down-conversion arranged in an array of length stride
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_rotation8_section3_avx512(double inp, const double* freq,
		double* out, const double* coeff, avx_rotation_struct* hist, size_t stride);


	/// Demodulation decimation on an array of input values using a single 
	/// down-conversion frequency. 
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded  second-order sections that form an IIR decimation filter. This 
	/// function has a restriction on the number of filter sections and input 
	/// strides, it only support strides of 8, 16, 24 or 32 and a filter of 
	/// order 6th (3 second order sections).
	/// 
	/// Selects AVX512 when available. If not, checks for AVX2. If neither, 
	/// uses the non-vectorized version.
	/// 
	/// @param inp Input values arranged as an array of length stride 
	/// @param freq Frequency of down-conversion
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_stride8_section3(const double* inp, double freq,
		double* out, const double* coeff, avx_stride_struct* hist, size_t stride);


        // Call during module initialization
        int demodulation_init(void);

	/// Demodulation decimation on an input value using multiple down-conversion
	/// frequencies.
	/// 
	/// Applies a demodulation in both cos and sin, then alculates a set of 
	/// cascaded  second-order sections that form an IIR decimation filter. This 
	/// function has a restriction on the number of filter sections and input 
	/// strides, it only support strides of 8, 16, 24 or 32 and a filter of 
	/// order 6th (3 second order sections).
	/// 
	/// Selects AVX512 when available. If not, checks for AVX2. If neither, 
	/// uses the non-vectorized version.
	/// 
	/// @param inp Input value
	/// @param freq Frequencies for down-conversion arranged in an array of length stride
	/// @param out Output values arranged as an array 2 x stride long
	/// @param coeff Filter coefficienst arranged in LIGO format (same for all filters)
	/// @param hist History and static values for this function 
	/// @param sections Number of second-order sections in the filter (3)
	/// @param stride Length of data array (8, 16, 24 or 32)
	void demodulation_decimation_rotation8_section3(double inp, const double* freq,
		double* out, const double* coeff, avx_rotation_struct* hist, size_t stride);


        /// A good set of decimation filter coefficients
        extern double section3_decim_coeffs[13];

#ifdef __cplusplus
}
#endif
/** @} */
